package be.iminds.ilabt.jfed.espec.bundle;

import be.iminds.ilabt.jfed.git.GitFetcher;
import be.iminds.ilabt.jfed.git.GitFetcherCommand;
import be.iminds.ilabt.jfed.util.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * BundleFetcher convenience factory to fetch ESpec Bundles in different ways.
 */
public class BundleFetcher {
    private static final Logger LOG = LoggerFactory.getLogger(BundleFetcher.class);

    private BundleFetcher() { }

    @Nonnull public static ESpecBundle fetchDir(@Nonnull String dirname) throws ESpecBundle.ESpecBundleInitException {
        return fetchDir(new File(dirname));
    }
    @Nonnull public static ESpecBundle fetchDir(@Nonnull File dir) throws ESpecBundle.ESpecBundleInitException {
        return new LocalDirESpecBundle(dir);
    }
    @Nonnull public static ESpecBundle fetchArchiveUrl(@Nonnull String url) throws ESpecBundle.ESpecBundleInitException, InterruptedException {
        try {
            return fetchArchiveUrl(new URL(url));
        } catch (MalformedURLException e) {
            throw new ESpecBundle.ESpecBundleInitException("Invalid URL", e);
        }
    }
    @Nonnull public static ESpecBundle fetchArchiveUrl(@Nonnull URL url) throws ESpecBundle.ESpecBundleInitException,InterruptedException {
        try {
            File tmpFile = IOUtils.urlToTmpFile(url);
            LOG.debug("Downloaded ESpecBundle to "+tmpFile.getPath());
            return fetchArchiveFile(tmpFile);
        } catch (IOException e) {
            throw new ESpecBundle.ESpecBundleInitException("Error downloading ESpecBundle", e);
        }
    }
    @Nonnull public static ESpecBundle fetchGitRepoDir(@Nonnull String gitFetcherSpec) throws ESpecBundle.ESpecBundleInitException,InterruptedException {
        try {
            GitFetcherCommand c = GitFetcherCommand.fromString(gitFetcherSpec)
                    .orElseThrow(() -> new IllegalArgumentException("gitFetcherSpec='"+gitFetcherSpec+"' is not a valid git source string"));
            File gitRepoDir = new GitFetcher().fetch(c);
            IOUtils.deleteDirAtExit(gitRepoDir);
            LOG.debug("Downloaded ESpec From repo to "+gitRepoDir);
            return fetchDir(gitRepoDir);
        } catch (Exception e) {
            throw new ESpecBundle.ESpecBundleInitException(
                    "Error downloading ESpecBundle from git repo '"+gitFetcherSpec+"'", e);
        }
    }
    @Nonnull public static ESpecBundle fetchArchiveFile(@Nonnull String filename) throws ESpecBundle.ESpecBundleInitException {
        return fetchArchiveFile(new File(filename));
    }
    @Nonnull public static ESpecBundle fetchArchiveFile(@Nonnull File filename) throws ESpecBundle.ESpecBundleInitException {
        return new LocalArchiveESpecBundle(filename);
    }
    @Nonnull public static ESpecBundle fetchArchive(@Nonnull InputStream inputStream) throws ESpecBundle.ESpecBundleInitException {
        return new InMemoryArchiveESpecBundle(inputStream);
    }
    @Nonnull public static ESpecBundle fetchArchive(@Nonnull byte[] bytes) throws ESpecBundle.ESpecBundleInitException {
        return new InMemoryArchiveESpecBundle(bytes);
    }
}
