package be.iminds.ilabt.jfed.espec.filefetcher;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.FileSource;
import be.iminds.ilabt.jfed.espec.model.GeneratedRSpecFileSource;
import be.iminds.ilabt.jfed.espec.model.GeneratedRandomFileSource;
import be.iminds.ilabt.jfed.espec.model.GitFileSource;
import be.iminds.ilabt.jfed.git.GitFetcher;
import be.iminds.ilabt.jfed.git.GitFetcherCommand;
import be.iminds.ilabt.jfed.git.GitFetcherCommandBuilder;
import be.iminds.ilabt.jfed.rspec.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.util.IOUtils;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;

/**
 * FileFetcher
 *
 * Fetches file content from whatever source is used in an experimentSpecification
 * See FileSource.SourceType for types
 *
 */
public class FileFetcher {
    @Nonnull private final FileSource fileSource;
    @Nonnull private final ESpecBundle bundle;
    @Nullable private final RSpecGeneratorFactory rSpecGeneratorFactory;

    public FileFetcher(@Nonnull FileSource fileSource,
                       @Nonnull ESpecBundle bundle,
                       @Nullable RSpecGeneratorFactory rSpecGeneratorFactory) {
        this.fileSource = fileSource;
        this.bundle = bundle;
        this.rSpecGeneratorFactory = rSpecGeneratorFactory;
    }

    @Nonnull
    public FileSource getFileSource() {
        return fileSource;
    }

    @Nonnull
    public ESpecBundle getBundle() {
        return bundle;
    }

    public boolean isFast() {
        switch (fileSource.getType()) {
            case BUNDLED: return true;
            case DIRECT: return true;
            case DOWNLOAD: return false;
            case GIT: return false;
            case GITHUB: return false;
            case META: return true;
            case GENERATED: return true;
            default: throw new IllegalStateException("Unsupported FileSource type: "+fileSource.getType().name());
        }
    }

    public boolean isDir() throws IOException {
        switch (fileSource.getType()) {
            case BUNDLED: return bundle.isDir(fileSource.getValue());
            case GIT: //fall-through
            case GITHUB: {
                GitFileSource gitFileSource = (GitFileSource) fileSource;
                return gitFileSource.isDir();
            }
            case DIRECT: //fall-through
            case DOWNLOAD: //fall-through
            default: return false;
        }
    }

    @Nonnull
    public byte[] fetchBytes() throws IOException {
        switch (fileSource.getType()) {
            case BUNDLED: return fetchBundled();
            case DIRECT: return fetchDirect();
            case DOWNLOAD: return fetchDownload();
            case GIT: return fetchGitFile();
            case GITHUB: return fetchGithubFile();
            case META: //nothing to do, meta files are only generated at the last moment
                throw new IllegalStateException("fetchBytes should not be called for META "+fileSource.getValue());
            case GENERATED: {
                switch (fileSource.getValue().toLowerCase()) {
                    case "random" : {
                        if (!(fileSource instanceof GeneratedRandomFileSource))
                            throw new IllegalStateException("Internal bug: random fileSource should be " +
                                    "instance of GeneratedRandomFileSource, not "+fileSource.getClass().getName());
                        return generateRandom((GeneratedRandomFileSource) fileSource);
                    }
                    case "rspec" : {
                        if (!(fileSource instanceof GeneratedRSpecFileSource))
                            throw new IllegalStateException("Internal bug: rspec fileSource should be " +
                                    "instance of GeneratedRSpecFileSource, not "+fileSource.getClass().getName());
                        return generateRspec((GeneratedRSpecFileSource) fileSource);
                    }
                    default: throw new IllegalStateException("fetchBytes should not be called for GENERATED "+fileSource.getValue());
                }
            }
            default: throw new IllegalStateException("Unsupported FileSource type: "+fileSource.getType().name());
        }
    }

    public byte[] generateRspec(GeneratedRSpecFileSource fileSource) {
        if (rSpecGeneratorFactory == null)
            throw new RuntimeException("Cannot generate RSpec: no rSpecGeneratorFactory registered");

        RSpecGenerator gen = rSpecGeneratorFactory.create();
        gen.setAm(fileSource.getAm());
        gen.setNodeCount(fileSource.getNodeCount());
        if (!fileSource.getNodeNames().isEmpty())
            gen.setNodeNames(fileSource.getNodeNames());
        if (!fileSource.getNodeNamePrefix().isEmpty())
            gen.setNodePrefix(fileSource.getNodeNamePrefix());
        if (fileSource.getResourceClass() != null)
            gen.setResourceClassIdPrefix(fileSource.getResourceClass());

        String rspec = gen.generateRequest();
        return rspec.getBytes(StandardCharsets.UTF_8);
    }
    public static byte[] generateRandom(GeneratedRandomFileSource fileSource) {
        String format = fileSource.getFormat();
        int len = fileSource.getLength();

        switch (format.trim().toLowerCase()) {
            case "password":
                return RandomDataGenerator.generatePassword(len).getBytes(StandardCharsets.UTF_8);
            case "binary":
                return RandomDataGenerator.generateBinary(len);
            case "base64":
                return RandomDataGenerator.generateBase64(len).getBytes(StandardCharsets.UTF_8);
            case "alphanum":
                return RandomDataGenerator.generateAlphaNum(len).getBytes(StandardCharsets.UTF_8);
            default:
                throw new IllegalStateException("generate:random does not support format \""+format+"\"");
        }
    }


    @Nonnull
    private byte[] fetchBundled() throws IOException {
        return bundle.getFileContent(fileSource.getValue());
    }

    @Nonnull
    private byte[] fetchDirect() {
        return fileSource.getValue().getBytes(StandardCharsets.UTF_8);
    }

    @Nonnull
    private byte[] fetchDownload() {
        try {
            URL url = new URL(fileSource.getValue());
            return IOUtils.urlToByteArray(url);
        } catch (MalformedURLException e) {
            throw new RuntimeException("Invalid URL", e);
        } catch (IOException e) {
            throw new RuntimeException("Error reading URL", e);
        }
    }

    @Nonnull
    private byte[] fetchGitFile() {
        GitFileSource gitFileSource = (GitFileSource) fileSource;
        assert gitFileSource.isFile();
        assert !gitFileSource.isDir();
        assert gitFileSource.getFilename() != null;
        if (!gitFileSource.isFile() || gitFileSource.getFilename() == null)
            throw new IllegalArgumentException("Cannot fetch bytes of a directory");

        GitFetcherCommand command = new GitFetcherCommandBuilder()
                .setGitUrl(gitFileSource.getGitUrl())
                .setBranch(gitFileSource.getBranch())
                .setRepoSubDir(gitFileSource.getRepoSubDir())
                .setUsername(gitFileSource.getUsername())
                .setPassword(gitFileSource.getPassword())
                .setPrivateKeyPem(gitFileSource.getPrivateKeyPem())
                .build();

        assert gitFileSource.isFile();
        assert !gitFileSource.isDir();
        assert gitFileSource.getFilename() != null;

        GitFetcher fetcher = new GitFetcher();
        File localDir = fetcher.fetch(command);
        File localFile = new File(localDir, gitFileSource.getFilename());
        if (!localFile.exists())
            throw new RuntimeException("Target file does not exist in downloaded git repo");
        try {
            return IOUtils.fileToByteArray(localFile);
        } catch (IOException e) {
            throw new RuntimeException("Error reading File downloaded from git", e);
        }
    }

    @Nonnull
    private byte[] fetchGithubFile() {
        //the same for now  (in future, support URLs to github page and handle that differently)
        return fetchGitFile();
    }
}
