package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.ExperimentSpecificationFileManager;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecConstants;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.sftp.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState does 3 things:
 *   - Setup SSH connections to nodes, for the dirs, upload, and execute steps
 *   - Do the "dir" step of the ExperimentSpecification
 *   - Do the "upload" step of the ExperimentSpecification
 *
 * Note that these 2 steps can be combined here easily, as there are no inter node dependencies.
 * For the execute step, things are different. See ExperimentSpecificationServicesSingleExecuteState
 *
 * Additionally, this class can be called to close all SSH connections. (typically after all executes)
 *   See TODO
 */
public class ExperimentSpecificationServicesDirsAndUploadsState
        extends SlicedState<ExperimentSpecificationServicesDirsAndUploadsState.ExperimentSpecificationServiceDirsAndUploadsStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesDirsAndUploadsState.class);
    @Nonnull
    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    @Nonnull
    private final Experiment experiment;

    @Nonnull
    private final ExperimentSpecification experimentSpecification;
    @Nonnull
    private final ESpecBundle experimentSpecificationBundle;
    @Nonnull
    private final ESpecLogger eSpecLogger;


    @Nonnull
    private final List<ExperimentSpecificationServiceDirsAndUploadsStateSlice> slices;

    private final ConcurrentHashMap<Pair<String, ESpecConstants.DirContent>, String> remotePathByDirClientIdAndDirContent = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Pair<String, UploadLikeSpec>, String> remotePathByClientIdAndUploadLikeSpec = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Pair<String, UploadLikeWithLogSpec>, String> remoteLogPathByClientIdAndUploadLikeWithLogSpec = new ConcurrentHashMap<>();

    protected ExperimentSpecificationServicesDirsAndUploadsState(@Nonnull SetupSoftwareExperimentJob job) {
        super("Execute ExperimentSpecification Dir And Upload Steps");
        this.setupSoftwareExperimentJob = job;
        this.experiment = job.getExperiment();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();

        experiment.requireSlice();
        if (experiment.getExperimentSpecification() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        assert experiment.getKeypairs() != null;

        this.slices = new ArrayList<>();


        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesDirsAndUploadsState will do nothing because there is no usable request RSpec");
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                //Note: Since this needs to setup SSH connections for execute as well, this will also run on nodes with only an execute step.

                boolean neededOnNode = experimentSpecification.getDirs().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, experimentSpecification, s));
                neededOnNode |= experimentSpecification.getExecutes().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, experimentSpecification, s));
                neededOnNode |= experimentSpecification.getUploads().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, experimentSpecification, s));

                //we assume that there is always an upload needed on the ansible node
                if (ESpecLogic.isAnsibleNode(nodeId, experimentSpecification))
                    neededOnNode = true;

                if (neededOnNode) {
                    slices.add(new ExperimentSpecificationServiceDirsAndUploadsStateSlice(job, node));
                }
            }
        }
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        ExperimentTaskStatus res = super.executeState(job);
        eSpecLogger.firePreExecute();
        return res;
    }


    @Override
    public Collection<ExperimentSpecificationServiceDirsAndUploadsStateSlice> getSlices() {
        return slices;
    }

    @Nullable
    public String getRemotePath(@Nonnull ESpecConstants.DirContent dirContent, @Nonnull String clientId) {
        return remotePathByDirClientIdAndDirContent.get(new Pair<>(clientId, dirContent));
    }
    @Nullable
    public String getRemotePath(@Nonnull UploadLikeSpec uploadLikeSpec, @Nonnull String clientId) {
        return remotePathByClientIdAndUploadLikeSpec.get(new Pair<>(clientId, uploadLikeSpec));
    }
    @Nullable
    public String getRemoteLogPath(@Nonnull UploadLikeWithLogSpec spec, @Nonnull String clientId) {
        return remoteLogPathByClientIdAndUploadLikeWithLogSpec.get(new Pair<>(clientId, spec));
    }


    public class AssureAllSshClosedStateSlice extends State {
        protected AssureAllSshClosedStateSlice() {
            super("Close All SSH Connections");
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            setupSoftwareExperimentJob.closeAllSshConnections();
            return ExperimentTaskStatus.SUCCESS;
        }
    }

    public AssureAllSshClosedStateSlice getAssureAllSshClosedStateSlice() {
        return new AssureAllSshClosedStateSlice();
    }

    public class ExperimentSpecificationServiceDirsAndUploadsStateSlice extends StateSlice {
        @Nonnull
        protected final FXRspecNode node;
        @Nonnull
        protected final String clientId;
        @Nullable
        private SSHClient ssh;

        @Nullable
        private ExternalFileUtil externalFileUtil;

        public ExperimentSpecificationServiceDirsAndUploadsStateSlice(@Nonnull Job<?> job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.node = node;
            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }


        private String dirPathToAbsolutePath(String dirPath) {
            if (externalFileUtil == null)
                throw new IllegalStateException("externalFileUtil == null");

            //3 cases for path:
            //  - absolute path (starts with /)   example: /work/data/ / /work /work/test/data
            //  - home path (starts with ~)       exammple: ~/data/  ~/   ~   ~/test/data
            //  - null -> home path               -> this is handled by the parser, so no need to check this here
            //TODO    empty path string or path not starting with / or ~  ->  error during parsing

            boolean absolutePath = dirPath.startsWith("/");
            assert absolutePath || dirPath.startsWith("~");

            if (absolutePath)
                return assureEndWithSlash(dirPath);

            String homeDir = assureEndWithSlash(externalFileUtil.getHomeDir());

            return assureEndWithSlash(homeDir + (dirPath.length() > 1 ? dirPath.substring(1) : ""));
        }

        @Override
        public String getName() {
            return "ExperimentSpecification steps for node " + node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExperimentSpecificationServiceDirsAndUploadsStateSlice.statefulRun due to SSH comnection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            externalFileUtil = new ExternalFileUtil(ssh);

            eSpecLogger.firePreDir();
            //TODO: only start when needed
            final SetupDirsState setupDirState = new SetupDirsState();
            setAndRunState(setupDirState);
            eSpecLogger.firePostDirAll(setupDirState.getStatus() == ExperimentTaskStatus.SUCCESS);

            if (setupDirState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;


            eSpecLogger.firePreUpload();
            final UploadState uploadState = new UploadState();
            try {
                setAndRunState(uploadState);
            } finally {
                eSpecLogger.firePostUploadAll(uploadState.getStatus() == ExperimentTaskStatus.SUCCESS);
            }

            if (uploadState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;

            return ExperimentTaskStatus.SUCCESS;
        }

        public class SetupDirsState extends State {
            protected SetupDirsState() {
                super("Checking (and possible creating) dirs on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                List<DirSpec> dirs = new ArrayList<>(experimentSpecification.getDirs());

                for (DirSpec dirSpec : dirs) {
                    if (ESpecLogic.mustRunOn(clientId, experimentSpecification, dirSpec)) {
                        String remotePath = dirPathToAbsolutePath(dirSpec.getPath());

                        int targetDirMask = ChModPermissionsUtil.toMask(dirSpec.getPermissions());
                        FileAttributes targetDirAtts = new FileAttributes.Builder()
                                .withPermissions(targetDirMask)
                                .withType(FileMode.Type.DIRECTORY).build();

                        if (dirSpec.getDirContent() != null) {
                            remotePathByDirClientIdAndDirContent.put(new Pair<>(clientId, dirSpec.getDirContent()), remotePath);
                        }

                        try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                            FileAttributes att = sftpClient.statExistence(remotePath);

                            LOG.debug("ExperimentSpecification.dir checked dir {} on {}. exists={}",
                                    remotePath, node.getClientId(), att != null);

                            if (att != null) {
                                //handle existing dir
                                if (att.getType() != FileMode.Type.DIRECTORY) {
                                    throw new IllegalStateException("Error while checking dir "+remotePath+", it already exists but is not a directory");
                                }

                                //don't check permissions, just set them
                                sftpClient.chmod(remotePath, targetDirMask);
                            } else {
                                //create dir(s)
                                SFTPEngine engine = sftpClient.getSFTPEngine();
                                final Deque<String> dirsToMake = new LinkedList<>();
                                for (PathComponents current = engine.getPathHelper().getComponents(remotePath); ;
                                     current = engine.getPathHelper().getComponents(current.getParent())) {
                                    final FileAttributes attrs = sftpClient.statExistence(current.getPath());
                                    if (attrs == null) {
                                        dirsToMake.push(current.getPath());
                                    } else if (attrs.getType() != FileMode.Type.DIRECTORY) {
                                        throw new IllegalStateException("Error while creating dir "+remotePath+" -> "+current.getPath() + " exists but is not a directory");
                                    } else {
                                        break;
                                    }
                                }
                                LOG.debug("ExperimentSpecification.dir must create dirs {}", dirsToMake);
                                while (!dirsToMake.isEmpty()) {
                                    String dirToMake = dirsToMake.pop();
                                    LOG.debug("ExperimentSpecification.dir creating {} on {}",
                                            dirToMake, node.getClientId());
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    engine.makeDir(dirToMake, targetDirAtts);
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    sftpClient.chmod(dirToMake, targetDirMask);
                                }
                            }

                            //TODO: we should set appropriate permissions along the parts of the directory path that already exist. ("execute" perm for uog depending if they get any permissions on final dir)

                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, true);
                        } catch (IOException ex) {
                            LOG.error("Error while creating dir {} on {}", remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while creating dir %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, false);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class UploadState extends State {
            protected UploadState() {
                super("Uploading files on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                //We need to wait for all files to be ready
                //noinspection ConstantConditions
                try {
                    job.getExperiment().getExperimentSpecificationFileManager().waitUntilReady();
                    if (job.getExperiment().getExperimentSpecificationFileManager().getErrorOccured())
                        throw new RuntimeException("Error fetching ESpec resources for upload");
                } catch (Exception e) {
                    LOG.debug("Error fetching ESpec resources for upload", e);
                    updateMessage("Error fetching ESpec resources for upload: "+e.getMessage());
                    return ExperimentTaskStatus.FAILED;
                }

                //uploads file AND scripts (but don't execute scripts yet)
                List<UploadLikeSpec> posUploads = new ArrayList<>(experimentSpecification.getUploads());
                posUploads.addAll(experimentSpecification.getExecutes());

                List<UploadLikeSpec> ansibleUploads = new ArrayList<>();
                if (experimentSpecification.getAnsible() != null) {
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getUploadLists());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getExecuteLists());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleGalaxySpecs());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsiblePlaybookSpecs());
                    for (AnsiblePlaybookSpec ansiblePlaybookSpec : experimentSpecification.getAnsible().getAnsiblePlaybookSpecs()) {
                        if (ansiblePlaybookSpec.hasExtraVarsJsonFileSpec() &&
                                ansiblePlaybookSpec.getExtraVarsJsonFileSpec() != null) {
                            ansibleUploads.add(ansiblePlaybookSpec.getExtraVarsJsonFileSpec());
                        }
                    }
                    posUploads.addAll(ansibleUploads);
                }

                List<UploadLikeSpec> localFiles = posUploads.stream()
                        .filter(f -> f.getSource() == null)
                        .collect(Collectors.toList());

                //exclude files that are already on the node
                List<UploadLikeSpec> uploads = posUploads.stream()
                        .filter(f -> f.getSource() != null)
                        .collect(Collectors.toList());


                LOG.debug("ExperimentSpecification.upload There are "+uploads.size()+" uploads needed.");
                LOG.debug("ExperimentSpecification.upload There are "+localFiles.size()+" files already present on the nodes.");
                LOG.debug("ExperimentSpecification.upload Of these there are "+ansibleUploads.size()+" ansible file uploads or already present files.");
                LOG.debug("ExperimentSpecification.upload Files to be uploaded: "+ uploads.stream()
                        .map(UploadLikeSpec::getDesc)
                        .collect(Collectors.joining(", ")));
                LOG.debug("ExperimentSpecification.upload Files already present: "+ localFiles.stream()
                        .map(UploadLikeSpec::getDesc)
                        .collect(Collectors.joining(", ")));

                String uploadDirSpecPath = null;
                String scriptDirSpecPath = null;
                String ansibleDirSpecPath = null;
                for (DirSpec dirSpec : experimentSpecification.getDirs()) {
                    if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                        if (dirSpec.isScriptContent()) {
                            scriptDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                        if (dirSpec.isUploadContent()) {
                            uploadDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                        if (dirSpec.isAnsibleContent()) {
                            ansibleDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                    }
                }
                assert scriptDirSpecPath != null;
                assert uploadDirSpecPath != null;
//                assert ansibleDirSpecPath != null; //can be null if no ansible


                //for local files, do not upload, but do register the remote path of the files
                for (UploadLikeSpec localFileSpec : localFiles) {
                    if (!ESpecLogic.mustRunOn(clientId, experimentSpecification, localFileSpec)) {
                        continue;
                    }

                    String remotePath = getRemotePath(externalFileUtil,
                            uploadDirSpecPath, scriptDirSpecPath, ansibleDirSpecPath,
                            localFileSpec);

                    remotePathByClientIdAndUploadLikeSpec.put(
                            new Pair<>(clientId, localFileSpec),
                            remotePath);

                    if (localFileSpec instanceof UploadLikeWithLogSpec) {
                        UploadLikeWithLogSpec es = (UploadLikeWithLogSpec) localFileSpec;
                        if (es.getLog() != null) {
                            String remoteLogPath;
                            if (!es.getLog().contains("/")) {
                                //file is considered to be in the same dir as the executed file
                                String dir = externalFileUtil.getPathOnly(remotePath);
                                remoteLogPath = assureEndWithSlash(dir)+es.getLog();
                            } else {
                                remoteLogPath = convertHomeDir(es.getLog(), externalFileUtil.getHomeDir());
                            }
                            assert remoteLogPath != null;
                            remoteLogPathByClientIdAndUploadLikeWithLogSpec.put(new Pair<>(clientId, es), remoteLogPath);
                        }
                    }
                }

                //upload the files that need to be uploaded, and register their remote path
                try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                    for (UploadLikeSpec uploadSpec : uploads) {
                        if (uploadSpec.getSource() == null) {
                            throw new RuntimeException("Internal assumption failed");
                        }

                        if (!ESpecLogic.mustRunOn(clientId, experimentSpecification, uploadSpec)) {
                            LOG.debug("ExperimentSpecification.upload skips {} on {}  (ansNode={} ansSub={})",
                                    uploadSpec.getDesc(),
                                    clientId,
                                    ESpecLogic.isAnsibleNode(clientId, experimentSpecification),
                                    ESpecLogic.isAnsibleSubPart(uploadSpec, experimentSpecification));
                            continue;
                        }

                        assert externalFileUtil != null;
                        //use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
                        String remotePath = getRemotePath(externalFileUtil,
                                uploadDirSpecPath, scriptDirSpecPath, ansibleDirSpecPath,
                                uploadSpec);

                        String name = externalFileUtil.getBasename(remotePath);

                        try {

                            boolean isDir = job.getExperiment().getExperimentSpecificationFileManager().isDir(uploadSpec.getSource());
                            if (isDir) {
                                if (!remotePath.endsWith("/"))
                                    remotePath += "/";

                                LOG.debug("ExperimentSpecification.upload starts recursive dir upload from {} to {} on {}",
                                        uploadSpec.getSource().getValue(), remotePath, node.getClientId());

                                Iterator<ExperimentSpecificationFileManager.DirFileBytes> filesToUpload =
                                        job.getExperiment().getExperimentSpecificationFileManager().getDirBytes(uploadSpec.getSource());
                                //note: filesToUpload will start with the dir itself (which is convenient)

                                while (filesToUpload.hasNext()) {
                                    ExperimentSpecificationFileManager.DirFileBytes dirFileBytes = filesToUpload.next();

                                    String curRemotePath = remotePath + dirFileBytes.getFullRelativePath();

                                    if (dirFileBytes.isDir()) {
                                        // We can't use uploadSpec.getPermissions()   here, as the parser takes a default,
                                        // but doesn't know at that point this is a dir
                                        // So we use fixed permissions for now. Not ideal there is no clean solution.
                                        // Users that wan't some fancy permissions for the dir, should chmod these themself
                                        makeDir(sftpClient, curRemotePath, "u=rwx");
                                    } else {
                                        //upload a file (directory already exists)
                                        uploadFile(sftpClient,
                                                curRemotePath,
                                                dirFileBytes.getBaseFilename().isEmpty() ?
                                                        name :
                                                        dirFileBytes.getBaseFilename(),
                                                uploadSpec.getPermissions(),
                                                dirFileBytes.getContent());
                                    }

                                    LOG.debug("ExperimentSpecification.upload finished recursive dir upload");
                                }
                            } else {
                                LOG.debug("ExperimentSpecification.upload starts single file upload from {} to {} on {}",
                                        uploadSpec.getSource().getValue(), remotePath, node.getClientId());

                                byte[] content;
                                if (uploadSpec.getSource().getType() == FileSource.SourceType.META) {
                                    content = getMetaContent(uploadSpec.getSource().getValue());
                                }else {
                                    if (uploadSpec.getSource().isGeneratedKeyPair()) {
                                        content = getGeneratedContent(uploadSpec.getSource().getValue());
                                    } else {
                                        content = job.getExperiment().getExperimentSpecificationFileManager().getBytes(uploadSpec.getSource());
                                    }
                                }

                                uploadFile(sftpClient,
                                        remotePath,
                                        name,
                                        uploadSpec.getPermissions(),
                                        content);
                            }

                            remotePathByClientIdAndUploadLikeSpec.put(new Pair<>(clientId, uploadSpec), remotePath);

                            if (uploadSpec instanceof UploadLikeWithLogSpec) {
                                UploadLikeWithLogSpec es = (UploadLikeWithLogSpec) uploadSpec;
                                if (es.getLog() != null) {
                                    String remoteLogPath;
                                    if (!es.getLog().contains("/")) {
                                        //file is considered to be in the same dir as the executed file
                                        String dir = externalFileUtil.getPathOnly(remotePath);
                                        remoteLogPath = assureEndWithSlash(dir)+es.getLog();
                                    } else {
                                        remoteLogPath = convertHomeDir(es.getLog(), externalFileUtil.getHomeDir());
                                    }
                                    assert remoteLogPath != null;
                                    remoteLogPathByClientIdAndUploadLikeWithLogSpec.put(new Pair<>(clientId, es), remoteLogPath);
                                }
                            }

                            eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, true);
                        } catch (IOException ex) {
                            LOG.error("Error while uploading {} to {} on {}", name, remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while uploading %s to %s on %s: %s", name, remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, false);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                } catch (IOException e) {
                    throw new JFedException("Failed to create SSH SFTP client", e);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            private void makeDir(SFTPClient sftpClient, String remotePath, String dirPerm) throws IOException {
                SFTPEngine engine = sftpClient.getSFTPEngine();

                int targetDirMask = ChModPermissionsUtil.toMask(dirPerm);
                FileAttributes targetDirAtts = new FileAttributes.Builder()
                        .withPermissions(targetDirMask)
                        .withType(FileMode.Type.DIRECTORY).build();

                LOG.debug("ExperimentSpecification.upload creating dir {} on {}",
                        remotePath, node.getClientId());
                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not always work...
                engine.makeDir(remotePath, targetDirAtts);
                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not always work...
                sftpClient.chmod(remotePath, targetDirMask);
            }

            private void uploadFile(SFTPClient sftpClient, String remotePath, String name, String perm, byte[] content) throws IOException {
                int targetFileMask = ChModPermissionsUtil.toMask(perm);

                LOG.debug("ExperimentSpecification.upload is uploading {} ({} bytes) to {} on {} with perm={}",
                        name, content.length, remotePath, node.getClientId(), perm);
                //TODO   NOTE: sftpClient.chmod and other permission related commands/parameters in sshj SFTP do not always work...
                sftpClient.getFileTransfer().upload(
                        new ByteArraySshjFile(name, content, perm),
                        remotePath);
                //TODO   NOTE: sftpClient.chmod and other permission related commands/parameters in sshj SFTP do not always work...
                //             u=rx seems to work, but oug=rw does not ?  (but needs testing)
                //             (also not know if when it works, if upload or chmod works or both)
                sftpClient.chmod(remotePath, targetFileMask);
            }



            @Nonnull
            private String makeRelativePathAbsolute(@Nonnull UploadLikeSpec uploadSpec,
                                                    @Nonnull ExternalFileUtil externalFileUtil,
                                                    @Nullable String uploadDirSpecPath,
                                                    @Nullable String scriptDirSpecPath,
                                                    @Nullable String ansibleDirSpecPath,
                                                    @Nonnull String path) {
                if (ESpecLogic.isAnsibleSubPart(uploadSpec, experimentSpecification)) {
                    if (ansibleDirSpecPath != null) {
                        return assureEndWithSlash(convertHomeDir(ansibleDirSpecPath, externalFileUtil.getHomeDir())) +
                                path;
                    } else {
                        return assureEndWithSlash(externalFileUtil.getHomeDir()) + "ansible/"+ path;
                    }
                } else if (uploadSpec instanceof UploadSpec) {
                    if (uploadDirSpecPath != null) {
                        return assureEndWithSlash(convertHomeDir(uploadDirSpecPath, externalFileUtil.getHomeDir())) +
                                path;
                    } else {
                        return assureEndWithSlash(externalFileUtil.getHomeDir()) + path;
                    }
                } else if (uploadSpec instanceof ExecuteSpec) {
                    if (scriptDirSpecPath != null) {
                        return assureEndWithSlash(convertHomeDir(scriptDirSpecPath, externalFileUtil.getHomeDir())) +
                                path;
                    } else {
                        return assureEndWithSlash(externalFileUtil.getHomeDir()) + path;
                    }
                } else {
                    throw new RuntimeException("BUG: uploadSpec of unknown class " + uploadSpec.getClass().getName());
                }
            }
            @Nonnull
            private String makeAnyPathAbsolute(@Nonnull UploadLikeSpec uploadSpec,
                                               @Nonnull ExternalFileUtil externalFileUtil,
                                               @Nullable String uploadDirSpecPath,
                                               @Nullable String scriptDirSpecPath,
                                               @Nullable String ansibleDirSpecPath,
                                               @Nonnull String path) {
                //3 cases for uploadSpec.getSource().getValue():
                //                                             (the file extensions in the examples below are not required at all)
                //  - absolute path (starts with /)   example: /work/data/filename.dat /testfile.txt
                //  - home path (starts with ~)       example: ~/data/filename.dat  ~/filename.txt
                //  - relative path                   example: filename.dat  test/filename.txt
                //       empty or null uploadSpec.getSource().getValue() -> error during parsing

                boolean absolutePath = path.startsWith("/");
                boolean homePath = path.startsWith("~");
                boolean relativePath = !(absolutePath || homePath);

                if (absolutePath) {
                    return path;
                }
                if (homePath) {
                    return convertHomeDir(path, externalFileUtil.getHomeDir());
                }

                assert relativePath;

                return makeRelativePathAbsolute(uploadSpec,
                        externalFileUtil,
                        uploadDirSpecPath,
                        scriptDirSpecPath,
                        ansibleDirSpecPath,
                        path);
            }

            @Nonnull
            private String getRemotePath(@Nonnull ExternalFileUtil externalFileUtil,
                                         @Nullable String uploadDirSpecPath,
                                         @Nullable String scriptDirSpecPath,
                                         @Nullable String ansibleDirSpecPath,
                                         @Nonnull UploadLikeSpec uploadSpec) {
                if (uploadSpec.getPath() == null) {
                    if (uploadSpec.getSource() == null)
                        throw new IllegalStateException("source and path should never both be null");

                    String sourceBasename = removeParentDirs(uploadSpec.getSource().getBasename());

                    return makeRelativePathAbsolute(uploadSpec,
                            externalFileUtil,
                            uploadDirSpecPath,
                            scriptDirSpecPath,
                            ansibleDirSpecPath,
                            sourceBasename);
                } else {
                    return makeAnyPathAbsolute(uploadSpec,
                            externalFileUtil,
                            uploadDirSpecPath,
                            scriptDirSpecPath,
                            ansibleDirSpecPath,
                            uploadSpec.getPath());
                }
            }

            private byte[] getGeneratedContent(String generatedContentId) {
                if (ESpecConstants.isPrivateKey(generatedContentId)) {
                    if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                        throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                    KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                    assert keyPair != null;
                    assert keyPair.getPrivate() != null;
                    char[] privateKeyChars = KeyUtil.privateKeyToPem(keyPair.getPrivate());
                    return new String(privateKeyChars).getBytes(StandardCharsets.UTF_8);
                }
                if (ESpecConstants.isPublicKey(generatedContentId)) {
                    if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                        throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                    KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                    assert keyPair != null;
                    assert keyPair.getPublic() != null;
                    String publicKeyString = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(keyPair.getPublic());
                    return publicKeyString.getBytes(StandardCharsets.UTF_8);
                }
                throw new UnsupportedOperationException("Unsupported \"generated\" source \""+generatedContentId+"\"");
            }

            private byte[] getMetaContent(String metaContentId) {
                if (metaContentId.equals(ESpecConstants.META_MANIFEST)) {
                    if (experiment.getSlice().getManifestRspec() == null)
                        throw new IllegalStateException("There is no manifest while uploading manifest.xml to node");
                    String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
                    if (manifestString == null)
                        throw new IllegalStateException("There is no manifest string while uploading manifest.xml to node");
                    return manifestString.getBytes(StandardCharsets.UTF_8);
                }
                throw new UnsupportedOperationException("Unsupported \"meta\" value \""+metaContentId+"\"");
            }
        }
    }

    @Nonnull
    private static String convertHomeDir(@Nonnull String path, @Nonnull String homeDirPath) {
        if (path.startsWith("~/") && path.length() > 2) return assureEndWithSlash(homeDirPath)+path.substring(2);
        if (path.startsWith("~") && path.length() > 1) return assureEndWithSlash(homeDirPath)+path.substring(1);
        if (path.equals("~")) return assureEndWithSlash(homeDirPath);
        return path;
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }

    @Nonnull
    private static String removeLastChildDir(@Nonnull String childDir, @Nonnull String path) {
        String childDirS = assureEndWithSlash(childDir);
        if (childDir.equals("/")) return path;
        assert path.endsWith(childDirS) : "\""+path+"\" does not end with \""+childDirS+"\"";
        return path.substring(0, path.length()-childDirS.length());
    }
    @Nonnull
    private static String removeParentDir(@Nonnull String parentPath, @Nonnull String path) {
        if (parentPath.isEmpty()) return path;
        assert path.startsWith(parentPath) : "\""+path+"\" does not start with \""+parentPath+"\"";
        return path.substring(parentPath.length());
    }
    @Nonnull
    private static String removeParentDirs(@Nonnull String path) {
        if (path.isEmpty() || path.indexOf('/') == -1 || path.length() <= 2) return path;

        int slash = path.lastIndexOf('/', path.length()-2);
        if (slash == -1)
            return path;
        return path.substring(slash+1, path.length());
    }
}
